from .type import Enum


class User(Enum):
    '''Enum for user data in login response'''
    LOGIN_NAME = "loginName"
    USER_TYPE = "userType"
    ID_CODE = "idCode"
    TITLE_TH = "titleTh"
    TITLE_EN = "titleEn"
    FIRST_NAME_TH = "firstNameTh"
    FIRST_NAME_EN = "firstNameEn"
    MIDDLE_NAME_TH = "middleNameTh"
    MIDDLE_NAME_EN = "middleNameEn"
    LAST_NAME_TH = "lastNameTh"
    LAST_NAME_EN = "lastNameEn"
    AVATAR = "avatar"
    GENDER = "gender"


class Student(Enum):
    '''Enum for student data in login response'''
    STD_ID = "stdId"
    STD_CODE = "stdCode"
    COPEN_ID = "copenId"
    COPEN_NAME_TH = "copenNameTh"
    COPEN_NAME_EN = "copenNameEn"
    CAMPUS_CODE = "campusCode"
    CAMPUS_NAME_TH = "campusNameTh"
    CAMPUS_NAME_EN = "campusNameEn"
    FACULTY_CODE = "facultyCode"
    FACULTY_NAME_TH = "facultyNameTh"
    FACULTY_NAME_EN = "facultyNameEn"
    DEPARTMENT_CODE = "departmentCode"
    DEPARTMENT_NAME_TH = "departmentNameTh"
    DEPARTMENT_NAME_EN = "departmentNameEn"
    MAJOR_CODE = "majorCode"
    MAJOR_NAME_TH = "majorNameTh"
    MAJOR_NAME_EN = "majorNameEn"
    NATION_CODE = "nationCode"
    NATIONALITY_NAME_TH = "nationalityNameTh"
    NATIONALITY_NAME_EN = "nationalityNameEn"
    STUDENT_STATUS_CODE = "studentStatusCode"
    STUDENT_STATUS_NAME_TH = "studentStatusNameTh"
    STUDENT_STATUS_NAME_EN = "studentStatusNameEn"
    STUDENT_TYPE_CODE = "studentTypeCode"
    STUDENT_TYPE_NAME_TH = "studentTypeNameTh"
    STUDENT_TYPE_NAME_EN = "studentTypeNameEn"
    EDULEVEL_CODE = "edulevelCode"
    EDULEVEL_NAME_TH = "edulevelNameTh"
    EDULEVEL_NAME_EN = "edulevelNameEn"
    STUDENT_YEAR = "studentYear"
    ADVISOR_ID = "advisorId"
    ADVISOR_NAME_TH = "advisorNameTh"
    ADVISOR_NAME_EN = "advisorNameEn"
    POSITION_TH = "positionTh"
    EMAIL = "email"
    MOBILE_NO = "mobileNo"


class Token(Enum):
    '''Enum for tokens in login response'''
    ACCESS_TOKEN = "accesstoken"
    RENEW_TOKEN = "renewtoken"

class Schedule(Enum):
    '''Enum for schedule data in schedule response'''
    ACADEMIC_YEAR = "academicYr"
    SEMESTER = "semester"

class FetchedResponses(Enum):
    '''Enum for fetched responses in client object'''
    LOGIN_RESPONSE = 0
    SCHEDULE_RESPONSE = 1

for enum in [User, Student, Token, Schedule, FetchedResponses]:
    for attr in enum:
        globals()[attr.name] = attr


from . import attribute, constant, url
from .type import ClientType, Dict, Enum, EnumMeta, Response, Union, Any


def response_to_json(response: Union[Response, dict]) -> dict:
    '''Convert response to json. Do nothing if response is already dictionary

    Parameters
    ----------
    response : (Response | dict)
        Response to convert

    Returns
    -------
        dict
    '''
    
    if isinstance(response, dict):
        return response
    
    if isinstance(response, Response):
        return response.json()

    raise TypeError("response must be Response or dict")


def extract(response: Union[Response, dict], attr: Enum) -> Any:
    '''Extract any value from login response or schedule response.
    Use enums from `pymyku.attribute` as key to get value.

    Parameters
    ----------
    attr : (Enum)

        Enum from `pymyku.attribute`.

    Returns
    -------
        Any    '''
    
    if isinstance(attr, EnumMeta):
        raise TypeError("attr must be Enum not EnumMeta. Use '.' operator after the EnumMeta object.")

    if not isinstance(attr, Enum):
        raise TypeError("attr must be Enum.")

    response = response_to_json(response)

    if isinstance(attr, attribute.User):
        return response.get("user", {})\
            .get(attr.value, None)

    if isinstance(attr, attribute.Student):
        return response.get("user", {})\
            .get("student", {})\
            .get(attr.value, None)

    if isinstance(attr, attribute.Token):
        return response.get(attr.value, None)

    if isinstance(attr, attribute.Schedule):
        result = response.get("results", [])
        if not result:
            return None
        return result[0].get(attr.value, None)

def get_raise(dict_: dict, key: str, dict_name:str = None) -> Any:
    '''Get value from dict and raise error if key is not found.

    Parameters
    ----------
    dict_ : dict
    key : str

    Returns
    -------
        Any

    Raises
    ------
        ValueError
    '''
    result = dict_.get(key)
    if result is None:
        raise ValueError(f'{key} is not found in {dict_name if dict_name else "dictionary"}')
    return result

def extract_user_data(login_response: Union[Response, dict]) -> dict:
    '''Extract user data from login response

    Parameters
    ----------
    login_response : (Response | dict)
        Response from the login request.

    Returns
    -------
        dict
    '''

    login_response = response_to_json(login_response)
        
    result = get_raise(login_response, 'user', 'login response')
    
    return result


def extract_student_data(login_response: Union[Response, dict]) -> dict:
    '''Extract student data from login response

    Parameters
    ----------
    login_response : (Response | dict)
        This is the response we get from the login request.

    Returns
    -------
        dict
    '''

    user_data = extract_user_data(login_response)
    
    result = get_raise(user_data, 'student', 'login response')    
    
    return result


def extract_access_token(login_response: Union[Response, dict]) -> str:
    '''Extract student code from login response

    Parameters
    ----------
    login_response : (Response | dict)
        This is the response from the login request.

    Returns
    -------
        str
    '''

    login_response = response_to_json(login_response)
    
    result = get_raise(login_response, 'accesstoken', 'login response')    
    
    return result


def extract_std_code(login_response: Union[Response, dict]) -> str:
    '''Extract student code from login response

    Parameters
    ----------
    login_response : (Response | dict)
        The response of the login request.

    Returns
    -------
        str
    '''

    login_response = response_to_json(login_response)

    student_data = extract_student_data(login_response)
    
    result = get_raise(student_data, 'stdCode', 'login response')

    return result


def extract_std_id(login_response: Union[Response, dict]) -> str:
    '''Extract student id from login response

    Parameters
    ----------
    login_response : (Response | dict)
        The response of the login request.

    Returns
    -------
        Student id.
    '''

    login_response = response_to_json(login_response)
    
    student_data = extract_student_data(login_response)
    
    result = get_raise(student_data, 'stdId', 'login response')
    
    return result


def extract_schedule(schedule_response: Union[Response, dict],
                     as_dict: bool = False,
                     full_result: bool = False) -> Union[tuple, dict, list]:
    '''Extract schedule (academic_year, semester) from schedule response

    Parameters
    ----------
    schedule_response : (Response | dict)
        The response from the schedule request.
    as_dict : (bool, optional)
        Whether to return the result as a dictionary or not. By default False
    full_result : (bool, optional)
        If True, return the full result (List[dict]), otherwise return the first item. By default False

    Returns
    -------
        tuple | dict | list
    '''

    schedule_response = response_to_json(schedule_response)
    
    result = get_raise(schedule_response, 'results', 'schedule response')
    
    if full_result:
        return result

    if isinstance(result, list):
        result = result[0]

    if as_dict:
        return result
    
    print(result)
    
    academic_year = get_raise(result, 'academicYr', 'schedule response')
    
    semester = get_raise(result, 'semester', 'schedule response')
    
    return academic_year, semester


def gen_request_headers(access_token: Union[str, Response, dict] = '') -> dict:
    '''Generate request headers.

    Parameters
    ----------
    access_token : (str | Response | dict, optional)
        MyKU acces token, by default None

    Returns
    -------
        dict
    '''

    header = {
        'app-key': constant.APP_KEY,
    }
    
    if isinstance(access_token, (Response, dict)):
        access_token = extract_access_token(access_token)
        
    if isinstance(access_token, str):
        header['x-access-token'] = access_token
    
    return header


def gen_login_request_params(username: str, password: str) -> dict:
    '''Generate request parameters for posting login request.

    Parameters
    ----------
    username : (str)
        Nontri account username (b##########)
    password : (str)
        Nontri account password (Don't worry, your password is not saved)

    Returns
    -------
        dict

    Format:
    ```python
        {
            'url': https://myapi.ku.th/auth/login,
            'data': {
                'username': username,
                'password': password,
            },
            'headers': {
                'app-key': constant.APP_KEY
        }
    }
    ```
    '''

    return {
        'url': url.LOGIN,
        'data': {
            'username': username,
            'password': password,
        },
        'headers': {
            'app-key': constant.APP_KEY
        }
    }


def __check_required_kwargs(kwargs: dict, required_kwargs: list) -> bool:

    for kwarg in required_kwargs:
        if not kwargs.get(kwarg) and not str(kwargs.get(kwarg)) == '0':
            raise ValueError(f'{kwarg} is required')
    return True


def gen_request_args_f(function: callable,
                         raise_exception: bool = True,
                         **kwargs) -> Dict[str, any]:
    '''Generate request parameters requied for the given function.

    Passing only `login_response`, `schedule_response` or `client` is also acceptable for some request parameters.

    Parameters
    ----------
    function : (callable)
        Request function from pymyku.request module.

    Returns
    -------
        Dict[str, any]
    '''
    
    if not callable(function):
        raise ValueError('function must be a callable')
        
    name = function.__name__

    if kwargs.get('client'):
        client = kwargs['client']
        kwargs['login_response'] = client.get(attribute.FetchedResponses.LOGIN_RESPONSE)
        kwargs['schedule_response'] = client.get(
            attribute.FetchedResponses.SCHEDULE_RESPONSE)

    if kwargs.get('login_response'):
        login_response = kwargs.get('login_response')
        login_response = response_to_json(login_response)

        user_data = extract_user_data(login_response)

        student_data = user_data['student']

        if not kwargs.get('access_token'):
            kwargs['access_token'] = extract_access_token(login_response)

        kwargs['user_type'] = user_data['userType']

        for key, value in constant.STUDENT_PARAM_DICT.items():
            if not kwargs.get(key):
                kwargs[key] = student_data.get(value)

    if kwargs.get('schedule_response'):
        schedule_response = kwargs.get('schedule_response')
        schedule_response = response_to_json(schedule_response)

        schedule = extract_schedule(schedule_response, as_dict=True)

        if not kwargs.get('academic_year'):
            kwargs['academic_year'] = schedule['academicYr']

        if not kwargs.get('semester'):
            kwargs['semester'] = schedule['semester']

    if raise_exception and not kwargs.get('access_token'):
        raise ValueError('access_token is required')

    headers = gen_request_headers(kwargs.get('access_token'))

    if name == 'logout':

        return {
            'url': url.LOGOUT,
            'headers': headers,
        }

    elif name == 'get_schedule':

        if raise_exception:
            __check_required_kwargs(kwargs, [
                'student_status_code', 'campus_code', 'faculty_code', 'major_code',
                'user_type'
            ])

        return {
            'url': url.SCHEDULE,
            'headers': headers,
            'params': {
                'stdStatusCode': kwargs.get('student_status_code'),
                'campusCode': kwargs.get('campus_code'),
                'facultyCode': kwargs.get('faculty_code'),
                'majorCode': kwargs.get('major_code'),
                'userType': kwargs.get('user_type')
            }
        }

    elif name == 'get_group_course':

        if raise_exception:
            __check_required_kwargs(kwargs, ['academic_year', 'semester', 'std_id'])

        return {
            'url': url.GROUP_COURSE,
            'headers': headers,
            'params': {
                "academicYear": kwargs.get('academic_year'),
                "semester": kwargs.get('semester'),
                "stdId": kwargs.get('std_id')
            }
        }
    elif name == 'get_check_grades':

        return {
            'url': url.CHECK_GRADES,
            'headers': headers,
            'params': {
                'stdCode': kwargs.get('std_code')
            }
        }

    elif name == 'get_gpax':

        return {
            'url': url.GPAX,
            'headers': headers,
            'params': {
                'stdId': kwargs.get('std_id')
            }
        }

    elif name == 'get_announce':

        if raise_exception:
            __check_required_kwargs(kwargs, ['std_id', 'academic_year', 'semester'])

        return {
            'url': url.ANNOUCE,
            'headers': headers,
            'params': {
                'stdId': kwargs.get('std_id'),
                'academicYear': kwargs.get('academic_year'),
                'semester': kwargs.get('semester')
            }
        }
    elif name == 'search_enroll':

        if raise_exception:
            __check_required_kwargs(kwargs, ['std_id', 'academic_year', 'semester'])

        return {
            'url': url.SEARCH_ENROLL,
            'headers': headers,
            'data': {
                'stdid': kwargs.get('std_id'),
                'academicYear': str(kwargs.get('academic_year')),
                'semester': str(kwargs.get('semester')),
            }
        }
    elif name == 'get_student_personal':

        if raise_exception:
            __check_required_kwargs(kwargs, ['std_id'])

        return {
            'url': url.STUDENT_PERSONAL,
            'headers': headers,
            'params': {
                'stdId': kwargs.get('std_id')
            }
        }
    elif name == 'get_student_education':

        if raise_exception:
            __check_required_kwargs(kwargs, ['std_id'])

        return {
            'url': url.STUDENT_EDUCATION,
            'headers': headers,
            'params': {
                'stdId': kwargs.get('std_id')
            }
        }
    elif name == 'get_student_address':

        if raise_exception:
            __check_required_kwargs(kwargs, ['std_id'])

        return {
            'url': url.STUDENT_ADDRESS,
            'headers': headers,
            'params': {
                'stdId': kwargs.get('std_id')
            }
        }
    elif name == 'search_subject':

        if raise_exception:
            __check_required_kwargs(kwargs, ['query'])

        return {
            'url': url.SEARCH_SUBJECT,
            'headers': headers,
            'params': {
                'query': kwargs.get('query'),
            }
        }
    elif name == 'search_subject_open':

        if raise_exception:
            __check_required_kwargs(kwargs,
                                    ['query', 'campus_code', 'academic_year', 'semester'])

        return {
            'url': url.SEARCH_SUBJECT_OPEN,
            'headers': headers,
            'params': {
                'query': kwargs.get('query'),
                'section': kwargs.get('section'),
                'campusCode': kwargs.get('campus_code'),
                'academicYear': kwargs.get('academic_year'),
                'semester': kwargs.get('semester')
            }
        }
    elif name == 'search_section_detail':

        if raise_exception:
            __check_required_kwargs(kwargs, ['section_id'])

        return {
            'url': url.SEARCH_SECTION_DETAIL,
            'headers': headers,
            'params': {
                'sectionId': kwargs.get('section_id')
            }
        }



from .attribute import FetchedResponses as FetchedResponsesAttr
from .attribute import Student as StudentAttr
from .attribute import Token as TokenAttr
from .attribute import User as UserAttr
from .attribute import Schedule as ScheduleAttr
from .constant import APP_KEY
from .exception import InvalidSubjectID, TokenExpired
from .pymyku import Client
from .type import ClientType, Response



from . import attribute, exception, requests, utils
from .type import Any, ClientType, Dict, Enum, EnumMeta, List, Response, Union


class Client(ClientType):
    '''    
    This class acts as a MyKu API wrapper.
    After created, the client will be logged in automatically.
    '''

    def __init__(self, username: str, password: str) -> None:
        self.__username: str = username
        self.__password: str = password

        self.initialize()

    def __valid_response(self,
                         response: Response,
                         to_json: bool = True) -> Union[dict, Response]:
        '''If the response is not 200, check if the error is due to an expired token. 
        If so, reset the token and raise an error. Otherwise, raise the error
        
        Parameters
        ----------
        response : (Response)
            The response object returned from the request.
        to_json : (bool, optional)
            If True, the response will be converted to a JSON object.
        
        Returns
        -------
            dict | Response 
        
        '''

        data = response.json()

        if response.status_code != 200:

            if data.get('code') == 'expired':
                self.reset()
                raise exception.TokenExpired(
                    "The access token has expired. Please reinitialize the client with .initialize()."
                )

        response.raise_for_status()

        return data if to_json else response

    def initialize(self) -> None:
        '''Initialize the client by logging in and fetch schedule data.
        
        Affected attribute
        -------------------
        `__login_response`,
        `__logged_in`,
        `__access_token`,
        `__schedule_response`,
        `__academic_year`,
        `__semester`

        '''

        self.login()

        self.__schedule_response = self.fetch_schedule()

        self.__academic_year, self.__semester = utils.extract_schedule(
            self.__schedule_response)

    def reset(self) -> None:
        '''Reset the client attribute to NoneType or Negative Value.

        Affected attribute
        -------------------
        `__login_response`,
        `__logged_in`,
        `__access_token`,
        `__schedule_response`,
        `__academic_year`,
        `__semester`
            
        '''
        self.__login_response = {}
        self.__logged_in = False
        self.__access_token = ''
        self.__schedule_response = {}
        self.__academic_year = None
        self.__semester = None

    def headers(self) -> dict:
        '''Return the headers for the requests.

        Returns
        -------
            utils.gen_request_headers(self.__access_token)
            
        '''
        return utils.gen_request_headers(self.__access_token)

    def login(self) -> Response:
        '''Login to MyKu and fetch the access token.

        Returns
        -------
            Response
            
        API
        ---
        https://myapi.ku.th/auth/login

        #
            
        '''

        login_response = requests.request_login(self.__username, self.__password)

        login_response.raise_for_status()

        self.__login_response = login_response.json()
        self.__logged_in = True
        self.__access_token = self.__login_response['accesstoken']

        return login_response

    def fetch_schedule(self, as_response: bool = False) -> Union[dict, Response]:
        '''Send GET request to MyKU common/getschedule API.

        Parameters
        ----------
        as_response : (bool, optional)
            Return as Response object if True, otherwise dict, by default False

        Returns
        -------
            dict | Response 
            
        API
        ---
        https://myapi.ku.th/common/getschedule

        #
        '''
        response = requests.get_schedule(login_response=self.__login_response)

        if as_response:
            return response

        return response.json()

    def fetch_group_course(self,
                           as_response: bool = False) -> Union[List[dict], Response]:
        '''Send GET request to MyKU std-profile/getGroupCourse API.

        Parameters
        ----------
        as_response : (bool, optional)
            Return as Response object if True, otherwise dict, by default False

        Returns
        -------
            List[dict] | Response 
        
        API
        ---
        https://myapi.ku.th/std-profile/getGroupCourse

        #
        '''

        response = requests.get_group_course(login_response=self.__login_response,
                                             schedule_response=self.__schedule_response)

        if as_response:
            return response

        return response.json()

    def fetch_grades(self, as_response: bool = False) -> List[dict]:
        '''Send GET request to MyKU std-profile/checkGrades API.

        Parameters
        ----------
        as_response : (bool, optional)
            Return as Response object if True, otherwise dict, by default False

        Returns
        -------
            dict | Response 

        API
        ---
        https://myapi.ku.th/std-profile/checkGrades

        #
        '''

        response = requests.get_check_grades(login_response=self.__login_response)

        if as_response:
            return response

        return response.json()

    def fetch_gpax(self, as_response: bool = False) -> dict:
        '''Send GET request to MyKU stddashboard/gpax API.
        
        Parameters
        ----------
        as_response : (bool, optional)
            Return as Response object if True, otherwise dict, by default False

        Returns
        -------
            dict | Response 

        API
        ---
        https://myapi.ku.th/stddashboard/gpax

        #
        '''

        response = requests.get_gpax(login_response=self.__login_response)

        if as_response:
            return response

        return response.json()

    def fetch_annouce(self,
                      academic_year=None,
                      semester=None,
                      as_response: bool = False) -> Union[List[dict], Response]:
        '''Send GET request to MyKU advisor/getAnnounceStd API.
                
        Parameters
        ----------
        as_response : (bool, optional)
            Return as Response object if True, otherwise dict, by default False

        Returns
        -------
            Union[List[dict], Response]


        API
        ---
        https://myapi.ku.th/advisor/getAnnounceStd

        #
        '''
        if academic_year is None:
            academic_year = self.__academic_year

        if semester is None:
            semester = self.__semester

        response = requests.get_announce(academic_year=academic_year,
                                         semester=semester,
                                         login_response=self.__login_response,
                                         schedule_response=self.__schedule_response)

        if as_response:
            return response

        return response.json()

    def search_enroll(self,
                      academic_year: Union[str, int] = None,
                      semester: Union[str, int] = None,
                      as_response: bool = False) -> Union[List[dict], Response]:
        '''Send GET request to MyKU enroll/searchEnrollResult API.

        Parameters
        ----------
        as_response : (bool, optional)
            Return as Response object if True, otherwise dict, by default False
            
        Returns
        -------
            Union[List[dict], Response]

        API
        ---
        https://myapi.ku.th/enroll/searchEnrollResult
        
        #
        '''

        if isinstance(semester, int):
            semester = str(semester)

        if isinstance(academic_year, int):
            academic_year = str(academic_year)

        response = requests.search_enroll(academic_year=academic_year,
                                          semester=semester,
                                          login_response=self.__login_response,
                                          schedule_response=self.__schedule_response)

        if as_response:
            return response

        return response.json()

    def fetch_student_personal(self, as_response: bool = False) -> Union[dict, Response]:
        '''Send GET request to MyKU std-profile/getStdPersonal API.

        Parameters
        ----------
        as_response : (bool, optional)
            Return as Response object if True, otherwise dict, by default False

        Returns
        -------
            dict | Response
    
        API
        ---
        https://myapi.ku.th/std-profile/getStdPersonal
        
        #
        '''

        response = requests.get_student_personal(login_response=self.__login_response)

        if as_response:
            return response

        return response.json()

    def fetch_student_education(self, as_response: bool = False) -> Union[dict, Response]:
        '''Send GET request to MyKU std-profile/getStdEducation API.

        Assigning only `login_response` or `client` is acceptable.

        Parameters
        ----------
        as_response : (bool, optional)
            Return as Response object if True, otherwise dict, by default False

        Returns
        -------
            dict | Response
        
        API
        ---
        https://myapi.ku.th/std-profile/getStdEducation
        
        #
        '''
        response = requests.get_student_education(login_response=self.__login_response)

        if as_response:
            return response

        return response.json()

    def fetch_student_address(self, as_response: bool = False) -> Union[dict, Response]:
        '''Send GET request to MyKU std-profile/getStdAddress API.

        Assigning only `login_response` or `client` is acceptable.

        Parameters
        ----------
        as_response : (bool, optional)
            Return as Response object if True, otherwise dict, by default False

        Returns
        -------
            dict | Response
        
        API
        ---
        https://myapi.ku.th/std-profile/getStdAddress
        
        #
        '''

        response = requests.get_student_address(login_response=self.__login_response)

        if as_response:
            return response

        return response.json()

    def search_subject_id(self, subject_id: str) -> List[Dict[str, str]]:
        '''Query subject with subject id by sending GET request to MyKU enroll/searchSubjectOpenEnr API.

        Parameters
        ----------
        subject_id : (str)
            Subject id to query. (At least 3 characters)
        
        Returns
        -------
            List[Dict[str, str]]
        
        API
        ---
        https://myapi.ku.th/enroll/searchSubjectOpenEnr
        
        #
        '''

        if len(subject_id) < 3:
            raise exception.InvalidSubjectID(
                "Subject ID must be at least 3 characters long.")

        response = requests.search_subject(query=subject_id,
                                           login_response=self.__login_response)

        response = self.__valid_response(response)

        return response.get('subjects', [])

    def search_subject_open(self, subject_id: str) -> List[Dict[str, Union[str, int]]]:
        '''Query subject enrollment info (All section) of current semester by sending GET request to MyKU enroll/openSubjectForEnroll API.

        Parameters
        ----------
        subject_id : (str)
            Subject id to query.

        Returns
        -------
            List[Dict[str, Union[str, int]]]
        
        API
        ---
        https://myapi.ku.th/enroll/openSubjectForEnroll
        
        #
        '''
        response = requests.search_subject_open(
            query=subject_id,
            login_response=self.__login_response,
            schedule_response=self.__schedule_response)

        response = self.__valid_response(response)

        return response.get('results', [])

    def get(self, attr: Enum) -> Any:
        '''Get any value from MyKU client. (login response and schedule response)
        Use enums from `pymyku.attribute` as key to get value.

        Parameters
        ----------
        attr : (Enum)
        
            Enum from `pymyku.attribute`.

        Returns
        -------
            Any
        '''

        if isinstance(attr, attribute.FetchedResponses):
            if attr.value == 0:
                return self.__login_response
            elif attr.value == 1:
                return self.__schedule_response

        if isinstance(attr, attribute.Schedule):
            return utils.extract(self.__schedule_response, attr)

        return utils.extract(self.__login_response, attr)

    def get_login_response(self) -> dict:
        '''Get login response from MyKU client.

        Returns
        -------
            dict
        '''
        return self.__login_response

    def get_user_data(self) -> dict:
        '''Get user data from MyKU client's login response.
        
        Returns
        -------
            dict
        '''
        login_response = self.get_login_response()
        return login_response.get('user', {})

    def get_student_data(self) -> dict:
        '''Get student data from MyKU client's login response.

        Returns
        -------
            dict
        '''
        user_data = self.get_user_data()
        return user_data.get('student', {})

    def get_schedule_response(self) -> dict:
        '''Get schedule response from MyKU client.
        Requires `fetch_schedule` to be called first.

        Returns
        -------
            dict
        '''

        return self.__schedule_response

    def get_access_token(self) -> str:
        '''Get access token from MyKU client.

        Returns
        -------
            str
        '''

        return self.__access_token

    def get_std_code(self) -> str:
        '''Get student code from MyKU client.

        Returns
        -------
            str
        '''
        return self.get(attribute.Student.STD_CODE)

    def get_group_course(self) -> List[dict]:
        '''Send GET request to MyKU std-profile/getGroupCourse API and return the result.

        Returns
        -------
            List[dict]
        '''
        response = self.fetch_group_course(as_response=True)

        response = self.__valid_response(response)

        return response.get('results', [])

    def get_gpax(self) -> float:
        '''Send GET request to MyKU std-profile/getGPAX API and return the gpax.

        Returns
        -------
            float
        '''

        response = self.fetch_gpax(as_response=True)

        response = self.__valid_response(response)

        return response.get('results', {})[0].get('gpax', {})

    def get_total_credit(self) -> int:
        '''Send GET request to MyKU std-profile/getTotalCredit API and return the total credit.

        Returns
        -------
            int
        '''

        response = self.fetch_gpax(as_response=True)

        response = self.__valid_response(response)

        return response.get('results', {})[0].get('total_credit', None)

    def get_grades(self, key="subject_code") -> Dict[str, Dict[str, str]]:
        '''Fetch grades for each subjects in all semesters.

        Parameters
        ----------
        key  : ((str, optional))
            Subject key from the response. 
            Can be `subject_code` or `subject_name_en`.
            Otherwise, `subject_code` will be used.
            Defaults to "subject_code".

        Returns
        -------
            Dict[str, Dict[str, str]]:
            
            {
                academic_year: {
                    semester: {
                        subject_code: grade
                        }
                }
            }
        '''
        grades = {}

        key = key if key in ("subject_code", "subject_name_en") else "subject_code"

        response = self.fetch_grades(as_response=True)

        response = self.__valid_response(response)

        for semester in response['results']:

            year = semester['academicYear']
            grades[year] = {}
            grades[year]['gpa'] = semester['gpa']
            grades[year]['cr'] = semester['cr']
            for subject in semester['grade']:
                grades[year][subject[key]] = subject['grade']

        return grades



LOGIN: str = 'https://myapi.ku.th/auth/login'
LOGOUT: str = 'https://myapi.ku.th/auth/logout'

GPAX: str = 'https://myapi.ku.th/stddashboard/gpax'

SCHEDULE: str = 'https://myapi.ku.th/common/getschedule'

ANNOUCE: str = 'https://myapi.ku.th/advisor/getAnnounceStd'

GROUP_COURSE: str = 'https://myapi.ku.th/std-profile/getGroupCourse'
CHECK_GRADES: str = 'https://myapi.ku.th/std-profile/checkGrades'
STUDENT_PERSONAL: str = 'https://myapi.ku.th/std-profile/getStdPersonal'
STUDENT_EDUCATION: str = 'https://myapi.ku.th/std-profile/getStdEducation'
STUDENT_ADDRESS: str = 'https://myapi.ku.th/std-profile/getStdAddress'

SEARCH_ENROLL: str = 'https://myapi.ku.th/enroll/searchEnrollResult'
SEARCH_SUBJECT: str = 'https://myapi.ku.th/enroll/searchSubjectOpenEnr'
SEARCH_SUBJECT_OPEN: str = 'https://myapi.ku.th/enroll/openSubjectForEnroll'
SEARCH_SECTION_DETAIL: str = 'https://myapi.ku.th/enroll/searchSectionDetail'



from setuptools import setup

setup(
    name='pymyku',
    version='0.1.3',
    description='Python MyKU API Wrapper',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='phusitsom',
    author_email='phusitsom@gmail.com',
    packages=['pymyku'],
    url='https://github.com/phusitsom/pymyku',
    install_requires=['requests'],
    license='GNU General Public License v3.0',
    python_requires='>=3.6',
    classifiers=[
        'Natural Language :: English',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ]
)



from requests import get as r_get
from requests import post as r_post

from . import utils
from .type import ClientType, Response, Union


def request_login(username: str, password: str) -> Response:
    '''Send POST request to MyKU auth/login API.

    Parameters
    ----------
    username : (str)
        Your Nontri account username (b##########)
    password : (str)
        Your password (Don't worry, your password is not saved)

    Returns
    -------
    Response

    API
    ---
    https://myapi.ku.th/auth/login

    #
    '''

    return r_post(**utils.gen_login_request_params(username, password))


def request_logout(access_token: str = '',
                   login_response: Union[Response, dict] = {},
                   client: ClientType = None) -> Response:
    '''Send POST request to MyKU auth/logout API.
    *I am uncertain that this request method will work properly.*

    Assigning only `login_response` or `client` is acceptable.

    Parameters
    ----------
    access_token : (str, optional)

    Returns
    -------
        Response

    API
    ---
    https://myapi.ku.th/auth/logout

    #
    '''

    params = utils.gen_request_args_f(logout, **locals())

    return r_get(**params)


def get_schedule(access_token: str = '',
                 user_type: str = '',
                 campus_code: str = '',
                 faculty_code: str = '',
                 major_code: str = '',
                 student_status_code: str = '',
                 login_response: Union[Response, dict] = {},
                 client: ClientType = None) -> Response:
    '''Send GET request to MyKU common/getschedule API.

    Assigning only `login_response` or `client` is acceptable.
    

    Parameters
    ----------
    access_token : (str, optional)
    user_type : (str, optional)
    campus_code : (str, optional)
    faculty_code : (str, optional)
    major_code : (str, optional)
    student_status_code : (dict, optional)

    Returns
    -------
    Response

    API
    ---
    https://myapi.ku.th/common/getschedule

    #
    '''
    params = utils.gen_request_args_f(get_schedule, **locals())

    return r_get(**params)


def get_group_course(access_token: str = '',
                     std_id: str = '',
                     academic_year: str = '',
                     semester: str = '',
                     login_response: Union[Response, dict] = {},
                     schedule_response: Union[Response, dict] = {},
                     client: ClientType = None) -> Response:
    '''Send GET request to MyKU std-profile/getGroupCourse API.

    Assigning only (`login_response` and `schedule_response`) or `client` is acceptable.
    
    Parameters
    ----------
    access_token : (str, optional)
    std_id : (str, optional)
    academic_year : (str, optional)
    semester : (str, optional)

    Returns
    -------
    Response

    API
    ---
    https://myapi.ku.th/std-profile/getGroupCourse

    #
    '''

    params = utils.gen_request_args_f(get_group_course, **locals())

    return r_get(**params)


def get_check_grades(access_token: str = '',
                     std_code: str = '',
                     login_response: Union[Response, dict] = {},
                     client: ClientType = None) -> Response:
    '''Send GET request to MyKU std-profile/checkGrades API.

    Assigning only `login_response` or `client` is acceptable.
    
    
    Parameters
    ----------
    access_token : (str, optional)
    std_code : (str, optional)

    Returns
    -------
    Response

    API
    ---
    https://myapi.ku.th/std-profile/checkGrades

    #
    '''

    params = utils.gen_request_args_f(get_check_grades, **locals())

    return r_get(**params)


def get_gpax(access_token: str = '',
             std_id: str = '',
             login_response: Union[Response, dict] = {},
             client: ClientType = None) -> Response:
    '''Send GET request to MyKU stddashboard/gpax API.
    
    Assigning only `login_response` or `client` is acceptable.

    Parameters
    ----------
    access_token : (_type_, optional)
    std_id : (str, optional)

    Returns
    -------
    Response

    API
    ---
    https://myapi.ku.th/stddashboard/gpax

    #
    '''

    params = utils.gen_request_args_f(get_gpax, **locals())

    return r_get(**params)


def get_announce(access_token: str = '',
                 std_id: str = '',
                 academic_year: str = '',
                 semester: str = '',
                 login_response: Union[Response, dict] = {},
                 schedule_response: Union[Response, dict] = {},
                 client: ClientType = None) -> Response:
    '''Send GET request to MyKU advisor/getAnnounceStd API.
    
    Assigning only (`login_response` and `schedule_response`) or `client` is acceptable.
    

    Parameters
    ----------
    access_token : (str, optional)
    std_id : (str, optional)
    academic_year : (str, optional)
    semester : (str, optional)

    Returns
    -------
    Response

    API
    ---
    https://myapi.ku.th/advisor/getAnnounceStd

    #
    '''

    params = utils.gen_request_args_f(get_announce, **locals())

    return r_get(**params)


def search_enroll(access_token: str = '',
                  std_id: str = '',
                  academic_year: str = '',
                  semester: str = '',
                  login_response: Union[Response, dict] = {},
                  schedule_response: Union[Response, dict] = {},
                  client: ClientType = None) -> Response:
    '''Send GET request to MyKU enroll/searchEnrollResult API.

    Assigning only (`login_response` and `schedule_response`) or `client` is acceptable.


    Parameters
    ----------
    access_token : (str, optional)
    std_id : (str, optional)
    academic_year : (str, optional)
    semester : (str, optional)

    Returns
    -------
    Response
    
    API
    ---
    https://myapi.ku.th/enroll/searchEnrollResult
    
    #    
    '''

    params = utils.gen_request_args_f(search_enroll, **locals())

    return r_post(**params)


def get_student_personal(access_token: str = '',
                         std_id: str = '',
                         login_response: Union[Response, dict] = {},
                         client: ClientType = None) -> Response:
    '''Send GET request to MyKU std-profile/getStdPersonal API.

    Assigning only `login_response` or `client` is acceptable.


    Parameters
    ----------
    access_token : (str, optional)
    std_id : (str, optional)

    Returns
    -------
    Response
    
    API
    ---
    https://myapi.ku.th/std-profile/getStdPersonal
    
    #    
    '''

    params = utils.gen_request_args_f(get_student_personal, **locals())

    return r_get(**params)


def get_student_education(access_token: str = '',
                          std_id: str = '',
                          login_response: Union[Response, dict] = {},
                          client: ClientType = None) -> Response:
    '''Send GET request to MyKU std-profile/getStdEducation API.

    Assigning only `login_response` or `client` is acceptable.

    Parameters
    ----------
    access_token : (str, optional)
    std_id : (str, optional)

    Returns
    -------
    Response
    
    API
    ---
    https://myapi.ku.th/std-profile/getStdEducation
    
    #    
    '''
    params = utils.gen_request_args_f(get_student_education, **locals())

    return r_get(**params)


def get_student_address(access_token: str = None,
                        std_id: str = '',
                        login_response: Union[Response, dict] = {},
                        client: ClientType = None) -> Response:
    '''Send GET request to MyKU std-profile/getStdAddress API.

    Assigning only `login_response` or `client` is acceptable.

    Parameters
    ----------
    access_token : (str, optional)
    std_id : (str, optional)

    Returns
    -------
    Response
    
    API
    ---
    https://myapi.ku.th/std-profile/getStdAddress
    
    #    
    '''

    params = utils.gen_request_args_f(get_student_address, **locals())

    return r_get(**params)


def search_subject(query: str,
                   access_token: str = '',
                   login_response: dict = {},
                   client: ClientType = None) -> Response:
    '''Send GET request to MyKU enroll/searchSubjectOpenEnr API.

    Assigning only `login_response` or `client` is acceptable.

    Parameters
    ----------
    query : (std)
    access_token : (str, optional)

    Returns
    -------
    Response
    
    API
    ---
    https://myapi.ku.th/enroll/searchSubjectOpenEnr
    
    #    
    '''

    params = utils.gen_request_args_f(search_subject, **locals())

    return r_get(**params)


def search_subject_open(query: str,
                        access_token: str = '',
                        section: str = '',
                        campus_code: str = '',
                        academic_year: str = '',
                        semester: str = '',
                        login_response: Union[Response, dict] = {},
                        schedule_response: Union[Response, dict] = {},
                        client: ClientType = None) -> Response:
    '''Send GET request to MyKU enroll/openSubjectForEnroll API.

    Assigning only (`login_response` and `schedule_response`) or `client` is acceptable.

    Parameters
    ----------
    query : (str)
    access_token : (str, optional)
    section : (str, optional)
    campus_code : (str, optional)
    academic_year : (str, optional)
    semester : (str, optional)

    Returns
    -------
    Response
    
    API
    ---
    https://myapi.ku.th/enroll/openSubjectForEnroll
    
    #    
    '''

    params = utils.gen_request_args_f(search_subject_open, **locals())

    return r_get(**params)


def search_section_detail(section_id: str,
                          access_token: str = '',
                          login_response: Union[Response, dict] = {},
                          client: ClientType = None) -> Response:
    '''Send GET request to MyKU enroll/searchSectionDetail API.

    Parameters
    ----------
    section_id : (str)
    access_token : (str, optional)

    Returns
    -------
    Response
    
    API
    ---
    https://myapi.ku.th/enroll/searchSectionDetail
    
    #    
    '''

    params = utils.gen_request_args_f(search_section_detail, **locals())

    return r_get(**params)



APP_KEY = 'txCR5732xYYWDGdd49M3R19o1OVwdRFc'

STUDENT_PARAM_NAMES = ('std_id', 'std_code', 'campus_code', 'faculty_code', 'major_code',
                       'student_status_code', 'student_year')

STUDENT_PARAM_RES_NAMES = ('stdId', 'stdCode', 'campusCode', 'facultyCode', 'majorCode',
                           'studentStatusCode', 'studentYear')

STUDENT_PARAM_DICT = dict(zip(STUDENT_PARAM_NAMES, STUDENT_PARAM_RES_NAMES))